@@ -14,6 +14,7 @@ import math
 import numbers
 import sys
 import array
+import codecs
 from collections import namedtuple
 from enum import Enum
 
@@ -28,45 +29,73 @@ try:
 except ImportError:
     from io import StringIO
 
-Config = namedtuple('Config', ('layers', 'codewords', 'cw_bits', 'bits', 'digits', 'text', 'bytes'))
+Config = namedtuple('Config', ('layers', 'codewords', 'cw_bits', 'bits'))
 
 configs = {
-    (15, True): Config(layers=1, codewords=17, cw_bits=6, bits=102, digits=13, text=12, bytes=6),
-    (19, False): Config(layers=1, codewords=21, cw_bits=6, bits=126, digits=18, text=15, bytes=8),
-    (19, True): Config(layers=2, codewords=40, cw_bits=6, bits=240, digits=40, text=33, bytes=19),
-    (23, False): Config(layers=2, codewords=48, cw_bits=6, bits=288, digits=49, text=40, bytes=24),
-    (23, True): Config(layers=3, codewords=51, cw_bits=8, bits=408, digits=70, text=57, bytes=33),
-    (27, False): Config(layers=3, codewords=60, cw_bits=8, bits=480, digits=84, text=68, bytes=40),
-    (27, True): Config(layers=4, codewords=76, cw_bits=8, bits=608, digits=110, text=89, bytes=53),
-    (31, False): Config(layers=4, codewords=88, cw_bits=8, bits=704, digits=128, text=104, bytes=62),
-    (37, False): Config(layers=5, codewords=120, cw_bits=8, bits=960, digits=178, text=144, bytes=87),
-    (41, False): Config(layers=6, codewords=156, cw_bits=8, bits=1248, digits=232, text=187, bytes=114),
-    (45, False): Config(layers=7, codewords=196, cw_bits=8, bits=1568, digits=294, text=236, bytes=145),
-    (49, False): Config(layers=8, codewords=240, cw_bits=8, bits=1920, digits=362, text=291, bytes=179),
-    (53, False): Config(layers=9, codewords=230, cw_bits=10, bits=2300, digits=433, text=348, bytes=214),
-    (57, False): Config(layers=10, codewords=272, cw_bits=10, bits=2720, digits=516, text=414, bytes=256),
-    (61, False): Config(layers=11, codewords=316, cw_bits=10, bits=3160, digits=601, text=482, bytes=298),
-    (67, False): Config(layers=12, codewords=364, cw_bits=10, bits=3640, digits=691, text=554, bytes=343),
-    (71, False): Config(layers=13, codewords=416, cw_bits=10, bits=4160, digits=793, text=636, bytes=394),
-    (75, False): Config(layers=14, codewords=470, cw_bits=10, bits=4700, digits=896, text=718, bytes=446),
-    (79, False): Config(layers=15, codewords=528, cw_bits=10, bits=5280, digits=1008, text=808, bytes=502),
-    (83, False): Config(layers=16, codewords=588, cw_bits=10, bits=5880, digits=1123, text=900, bytes=559),
-    (87, False): Config(layers=17, codewords=652, cw_bits=10, bits=6520, digits=1246, text=998, bytes=621),
-    (91, False): Config(layers=18, codewords=720, cw_bits=10, bits=7200, digits=1378, text=1104, bytes=687),
-    (95, False): Config(layers=19, codewords=790, cw_bits=10, bits=7900, digits=1511, text=1210, bytes=753),
-    (101, False): Config(layers=20, codewords=864, cw_bits=10, bits=8640, digits=1653, text=1324, bytes=824),
-    (105, False): Config(layers=21, codewords=940, cw_bits=10, bits=9400, digits=1801, text=1442, bytes=898),
-    (109, False): Config(layers=22, codewords=1020, cw_bits=10, bits=10200, digits=1956, text=1566, bytes=976),
-    (113, False): Config(layers=23, codewords=920, cw_bits=12, bits=11040, digits=2116, text=1694, bytes=1056),
-    (117, False): Config(layers=24, codewords=992, cw_bits=12, bits=11904, digits=2281, text=1826, bytes=1138),
-    (121, False): Config(layers=25, codewords=1066, cw_bits=12, bits=12792, digits=2452, text=1963, bytes=1224),
-    (125, False): Config(layers=26, codewords=1144, cw_bits=12, bits=13728, digits=2632, text=2107, bytes=1314),
-    (131, False): Config(layers=27, codewords=1224, cw_bits=12, bits=14688, digits=2818, text=2256, bytes=1407),
-    (135, False): Config(layers=28, codewords=1306, cw_bits=12, bits=15672, digits=3007, text=2407, bytes=1501),
-    (139, False): Config(layers=29, codewords=1392, cw_bits=12, bits=16704, digits=3205, text=2565, bytes=1600),
-    (143, False): Config(layers=30, codewords=1480, cw_bits=12, bits=17760, digits=3409, text=2728, bytes=1702),
-    (147, False): Config(layers=31, codewords=1570, cw_bits=12, bits=18840, digits=3616, text=2894, bytes=1806),
-    (151, False): Config(layers=32, codewords=1664, cw_bits=12, bits=19968, digits=3832, text=3067, bytes=1914),
+    (15, True): Config(layers=1, codewords=17, cw_bits=6, bits=102),
+    (19, False): Config(layers=1, codewords=21, cw_bits=6, bits=126),
+    (19, True): Config(layers=2, codewords=40, cw_bits=6, bits=240),
+    (23, False): Config(layers=2, codewords=48, cw_bits=6, bits=288),
+    (23, True): Config(layers=3, codewords=51, cw_bits=8, bits=408),
+    (27, False): Config(layers=3, codewords=60, cw_bits=8, bits=480),
+    (27, True): Config(layers=4, codewords=76, cw_bits=8, bits=608),
+    (31, False): Config(layers=4, codewords=88, cw_bits=8, bits=704),
+    (37, False): Config(layers=5, codewords=120, cw_bits=8, bits=960),
+    (41, False): Config(layers=6, codewords=156, cw_bits=8, bits=1248),
+    (45, False): Config(layers=7, codewords=196, cw_bits=8, bits=1568),
+    (49, False): Config(layers=8, codewords=240, cw_bits=8, bits=1920),
+    (53, False): Config(layers=9, codewords=230, cw_bits=10, bits=2300),
+    (57, False): Config(layers=10, codewords=272, cw_bits=10, bits=2720),
+    (61, False): Config(layers=11, codewords=316, cw_bits=10, bits=3160),
+    (67, False): Config(layers=12, codewords=364, cw_bits=10, bits=3640),
+    (71, False): Config(layers=13, codewords=416, cw_bits=10, bits=4160),
+    (75, False): Config(layers=14, codewords=470, cw_bits=10, bits=4700),
+    (79, False): Config(layers=15, codewords=528, cw_bits=10, bits=5280),
+    (83, False): Config(layers=16, codewords=588, cw_bits=10, bits=5880),
+    (87, False): Config(layers=17, codewords=652, cw_bits=10, bits=6520),
+    (91, False): Config(layers=18, codewords=720, cw_bits=10, bits=7200),
+    (95, False): Config(layers=19, codewords=790, cw_bits=10, bits=7900),
+    (101, False): Config(layers=20, codewords=864, cw_bits=10, bits=8640),
+    (105, False): Config(layers=21, codewords=940, cw_bits=10, bits=9400),
+    (109, False): Config(layers=22, codewords=1020, cw_bits=10, bits=10200),
+    (113, False): Config(layers=23, codewords=920, cw_bits=12, bits=11040),
+    (117, False): Config(layers=24, codewords=992, cw_bits=12, bits=11904),
+    (121, False): Config(layers=25, codewords=1066, cw_bits=12, bits=12792),
+    (125, False): Config(layers=26, codewords=1144, cw_bits=12, bits=13728),
+    (131, False): Config(layers=27, codewords=1224, cw_bits=12, bits=14688),
+    (135, False): Config(layers=28, codewords=1306, cw_bits=12, bits=15672),
+    (139, False): Config(layers=29, codewords=1392, cw_bits=12, bits=16704),
+    (143, False): Config(layers=30, codewords=1480, cw_bits=12, bits=17760),
+    (147, False): Config(layers=31, codewords=1570, cw_bits=12, bits=18840),
+    (151, False): Config(layers=32, codewords=1664, cw_bits=12, bits=19968),
+}
+
+encoding_to_eci = {
+    'cp437': 0, # also 2
+    'iso8859-1': 1, # (also 3) default interpretation, readers should assume if no ECI mark
+    'iso8859-2': 4,
+    'iso8859-3': 5,
+    'iso8859-4': 6,
+    'iso8859-5': 7,
+    'iso8859-6': 8,
+    'iso8859-7': 9,
+    'iso8859-8': 10,
+    'iso8859-9': 11,
+    'iso8859-13': 15,
+    'iso8859-14': 16,
+    'iso8859-15': 17,
+    'iso8859-16': 18,
+    'shift_jis': 20,
+    'cp1250': 21,
+    'cp1251': 22,
+    'cp1252': 23,
+    'cp1256': 24,
+    'utf-16-be': 25, # no BOM
+    'utf-8': 26,
+    'ascii': 27, # also 170
+    'big5': 28,
+    'gb18030': 29,
+    'euc_kr': 30,
 }
 
 polynomials = {
@@ -77,6 +106,8 @@ polynomials = {
     12: 4201,
 }
 
+Side = Enum('Side', ('left', 'right', 'bottom', 'top'))
+
 Mode = Enum('Mode', ('UPPER', 'LOWER', 'MIXED', 'PUNCT', 'DIGIT', 'BINARY'))
 Latch = Enum('Latch', Mode.__members__)
 Shift = Enum('Shift', Mode.__members__)
@@ -99,7 +130,7 @@ latch_len = {
         Mode.UPPER: 0, Mode.LOWER: 5, Mode.MIXED: 5, Mode.PUNCT: 10, Mode.DIGIT: 5, Mode.BINARY: 10
     },
     Mode.LOWER: {
-        Mode.UPPER: 10, Mode.LOWER: 0, Mode.MIXED: 5, Mode.PUNCT: 10, Mode.DIGIT: 5, Mode.BINARY: 10
+        Mode.UPPER: 9, Mode.LOWER: 0, Mode.MIXED: 5, Mode.PUNCT: 10, Mode.DIGIT: 5, Mode.BINARY: 10
     },
     Mode.MIXED: {
         Mode.UPPER: 5, Mode.LOWER: 5, Mode.MIXED: 0, Mode.PUNCT: 5, Mode.DIGIT: 10, Mode.BINARY: 10
@@ -116,24 +147,12 @@ latch_len = {
 }
 
 shift_len = {
-    Mode.UPPER: {
-        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
-    },
-    Mode.LOWER: {
-        Mode.UPPER: 5, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
-    },
-    Mode.MIXED: {
-        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
-    },
-    Mode.PUNCT: {
-        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: E, Mode.DIGIT: E, Mode.BINARY: E
-    },
-    Mode.DIGIT: {
-        Mode.UPPER: 4, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 4, Mode.DIGIT: E, Mode.BINARY: E
-    },
-    Mode.BINARY: {
-        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: E, Mode.DIGIT: E, Mode.BINARY: E
-    },
+    (Mode.UPPER, Mode.PUNCT): 5,
+    (Mode.LOWER, Mode.UPPER): 5,
+    (Mode.LOWER, Mode.PUNCT): 5,
+    (Mode.MIXED, Mode.PUNCT): 5,
+    (Mode.DIGIT, Mode.UPPER): 4,
+    (Mode.DIGIT, Mode.PUNCT): 4,
 }
 
 char_size = {
@@ -194,20 +213,30 @@ def reed_solomon(wd, nd, nc, gf, pp):
                 wd[nd + j] ^= wd[nd + j + 1]
 
 
-def find_optimal_sequence(data):
+def find_optimal_sequence(data, encoding=None):
     """ Find optimal sequence, i.e. with minimum number of bits to encode data.
 
     TODO: add support of FLG(n) processing
 
     :param data: string or bytes to encode
+    :param encoding: see :py:class:`AztecCode`
     :return: optimal sequence
     """
+
+    # standardize encoding name, ensure that it's valid for ECI, and encode string to bytes
+    if encoding:
+        encoding = codecs.lookup(encoding).name
+        eci = encoding_to_eci[encoding]
+    else:
+        encoding = 'iso8859-1'
+        eci = None
+    if isinstance(data, str):
+        data = data.encode(encoding)
+
     back_to = {m: Mode.UPPER for m in Mode}
     cur_len = {m: 0 if m==Mode.UPPER else E for m in Mode}
     cur_seq = {m: [] for m in Mode}
     prev_c = None
-    if isinstance(data, str):
-        data = data.encode('iso-8859-1')
     for c in data:
         for x in Mode:
             for y in Mode:
@@ -232,6 +261,8 @@ def find_optimal_sequence(data):
                             cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch.MIXED, Latch.PUNCT]
                         elif x in (Mode.PUNCT, Mode.DIGIT) and y != Mode.UPPER:
                             cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch[y.name]]
+                        elif x == Mode.LOWER and y == Mode.UPPER:
+                            cur_seq[y] += [Latch.DIGIT, Latch.UPPER]
                         elif x in (Mode.UPPER, Mode.LOWER) and y == Mode.PUNCT:
                             cur_seq[y] += [Latch.MIXED, Latch[y.name]]
                         elif x == Mode.MIXED and y != Mode.UPPER:
@@ -250,7 +281,7 @@ def find_optimal_sequence(data):
                                 cur_seq[y] += [Misc.RESUME]
                             elif y == Mode.UPPER:
                                 if back_to[x] == Mode.LOWER:
-                                    cur_seq[y] += [Misc.RESUME, Latch.MIXED, Latch.UPPER]
+                                    cur_seq[y] += [Misc.RESUME, Latch.DIGIT, Latch.UPPER]
                                 if back_to[x] == Mode.MIXED:
                                     cur_seq[y] += [Misc.RESUME, Latch.UPPER]
                             elif y == Mode.LOWER:
@@ -274,6 +305,8 @@ def find_optimal_sequence(data):
                         # TODO: update for digit
                         if x in (Mode.PUNCT, Mode.DIGIT):
                             cur_seq[y] = [Latch.UPPER, Latch[y.name]]
+                        elif x == Mode.LOWER and y == Mode.UPPER:
+                            cur_seq[y] = [Latch.DIGIT, Latch.UPPER]
                         elif x in (Mode.BINARY, Mode.UPPER, Mode.LOWER) and y == Mode.PUNCT:
                             cur_seq[y] = [Latch.MIXED, Latch[y.name]]
                         else:
@@ -292,16 +325,14 @@ def find_optimal_sequence(data):
                 next_len[x] = cur_len[x] + char_size[x]
                 next_seq[x] = cur_seq[x] + [c]
             for y in Mode:
-                if y == x or y == Mode.BINARY:
-                    continue
-                if cur_len[y] + shift_len[y][x] + char_size[x] < next_len[y]:
-                    next_len[y] = cur_len[y] + shift_len[y][x] + char_size[x]
+                if (y, x) in shift_len and cur_len[y] + shift_len[(y, x)] + char_size[x] < next_len[y]:
+                    next_len[y] = cur_len[y] + shift_len[y, x] + char_size[x]
                     next_seq[y] = cur_seq[y] + [Shift[x.name], c]
         # TODO: review this!!!
         if prev_c and bytes((prev_c, c)) in punct_2_chars:
             for x in Mode:
                 # Will never StopIteration because we must have one S/L already since prev_c is PUNCT
-                last_mode = next(s.value for s in reversed(cur_seq[x]) if s in Latch or s in Shift)
+                last_mode = next(s.value for s in reversed(cur_seq[x]) if isinstance(s, Latch) or isinstance(s, Shift))
                 if last_mode == Mode.PUNCT:
                     last_c = cur_seq[x][-1]
                     if isinstance(last_c, int) and bytes((last_c, c)) in punct_2_chars:
@@ -350,6 +381,9 @@ def find_optimal_sequence(data):
         if c == Shift.BINARY:
             is_binary_length = True
 
+    if eci is not None:
+        updated_result_seq = [ Shift.PUNCT, Misc.FLG, len(str(eci)), eci ] + updated_result_seq
+
     return updated_result_seq
 
 
@@ -373,7 +407,7 @@ def optimal_sequence_to_bits(optimal_sequence):
             mode = prev_mode
             shift = False
         # get mode from sequence character
-        if ch in Latch:
+        if isinstance(ch, Latch):
             mode = ch.value
         # handle FLG(n)
         elif ch == Misc.FLG:
@@ -393,7 +427,7 @@ def optimal_sequence_to_bits(optimal_sequence):
                 eci_code = sequence.pop(0)
                 if not isinstance(eci_code, numbers.Number) or not 0 <= eci_code < (10**flg_n):
                     raise Exception('Expected FLG({}) ECI code to be a number from 0 to {}'.format(flg_n, (10**flg_n) - 1))
-                out_digits = b'%.*d' % (flg_n, eci_code)
+                out_digits = str(eci_code).zfill(flg_n).encode()
                 for ch in out_digits:
                     index = code_chars[Mode.DIGIT].index(ch)
                     out_bits += bin(index)[2:].zfill(char_size[Mode.DIGIT])
@@ -417,7 +451,7 @@ def optimal_sequence_to_bits(optimal_sequence):
                 ch = sequence.pop(0)
                 out_bits += bin(ch)[2:].zfill(char_size[Mode.BINARY])
         # handle other shift
-        elif ch in Shift:
+        elif isinstance(ch, Shift):
             mode, prev_mode = ch.value, mode
             shift = True
     return out_bits
@@ -467,17 +501,18 @@ def get_config_from_table(size, compact):
     except KeyError:
         raise NotImplementedError('Failed to find config with size and compactness flag')
 
-def find_suitable_matrix_size(data, ec_percent=23):
+def find_suitable_matrix_size(data, ec_percent=23, encoding=None):
     """ Find suitable matrix size
     Raise an exception if suitable size is not found
 
     :param data: string or bytes to encode
     :param ec_percent: percentage of symbol capacity for error correction (default 23%)
+    :param encoding: see :py:class:`AztecCode`
     :return: (size, compact) tuple
     """
-    optimal_sequence = find_optimal_sequence(data)
+    optimal_sequence = find_optimal_sequence(data, encoding)
     out_bits = optimal_sequence_to_bits(optimal_sequence)
-    for (size, compact) in configs.keys():
+    for (size, compact) in sorted(configs.keys()):
         config = get_config_from_table(size, compact)
         bits = config.bits
         # calculate minimum required number of bits
@@ -491,7 +526,7 @@ class AztecCode(object):
     Aztec code generator
     """
 
-    def __init__(self, data, size=None, compact=None, ec_percent=23):
+    def __init__(self, data, size=None, compact=None, ec_percent=23, encoding=None):
         """ Create Aztec code with given data.
         If size and compact parameters are None (by default), an
         optimal size and compactness calculated based on the data.
@@ -500,18 +535,22 @@ class AztecCode(object):
         :param size: size of matrix
         :param compact: compactness flag
         :param ec_percent: percentage of symbol capacity for error correction (default 23%)
+        :param encoding:
+          If set, sequence will include an initial ECI mark corresponding to the specified encoding (see :py:mod:`codecs`)
+          If unset, no ECI mark will be included and string must be encodable as 'iso8859-1'
         """
         self.data = data
+        self.encoding = encoding
         self.sequence = None
         self.ec_percent = ec_percent
         if size is not None and compact is not None:
-            if (size, compact) in table:
+            if (size, compact) in configs:
                 self.size, self.compact = size, compact
             else:
                 raise Exception(
                     'Given size and compact values (%s, %s) are not found in sizes table!' % (size, compact))
         else:
-            self.size, self.compact, self.sequence = find_suitable_matrix_size(self.data, ec_percent)
+            self.size, self.compact, self.sequence = find_suitable_matrix_size(self.data, ec_percent, encoding)
         self.__create_matrix()
         self.__encode_data()
 
@@ -551,11 +590,10 @@ class AztecCode(object):
 
     def print_out(self, border=0):
         """ Print out Aztec code matrix using ASCII output """
-        if not fancy:
-            print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
-            for line in self.matrix:
-                print(' '*border + ''.join(('#' if x else ' ') for x in line) + ' '*border)
-            print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
+        print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
+        for line in self.matrix:
+            print(' '*border + ''.join(('#' if x else ' ') for x in line) + ' '*border)
+        print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
 
     def print_fancy(self, border=0):
         """ Print out Aztec code matrix using Unicode box-drawing characters and ANSI colorization """
@@ -680,14 +718,15 @@ class AztecCode(object):
             self.matrix[center + y][center + x] = (bit == '1')
             index += 1
 
-    def __add_data(self, data):
+    def __add_data(self, data, encoding):
         """ Add data to encode to the matrix
 
         :param data: data to encode
+        :param encoding: see :py:class:`AztecCode`
         :return: number of data codewords
         """
         if not self.sequence:
-            self.sequence = find_optimal_sequence(data)
+            self.sequence = find_optimal_sequence(data, encoding)
         out_bits = optimal_sequence_to_bits(self.sequence)
         config = get_config_from_table(self.size, self.compact)
         layers_count = config.layers
@@ -695,11 +734,8 @@ class AztecCode(object):
         cw_bits = config.cw_bits
         bits = config.bits
 
-        # error correction percent
-        ec_percent = 23  # recommended
         # calculate minimum required number of bits
-        required_bits_count = int(math.ceil(len(out_bits) * 100.0 / (
-            100 - ec_percent) + 3 * 100.0 / (100 - ec_percent)))
+        required_bits_count = int(math.ceil((len(out_bits) + 3) * 100.0 / (100 - self.ec_percent)))
         data_codewords = get_data_codewords(out_bits, cw_bits)
         if required_bits_count > bits:
             raise Exception('Data too big to fit in Aztec code with current size!')
@@ -713,7 +749,7 @@ class AztecCode(object):
         ring_radius = 5 if self.compact else 7
 
         num = 2
-        side = 'top'
+        side = Side.top
         layer_index = 0
         pos_x = center - ring_radius
         pos_y = center - ring_radius - 1
@@ -724,7 +760,7 @@ class AztecCode(object):
             bits_pair = [(bit == '1') for bit in full_bits[i:i + 2]]
             if layer_index >= layers_count:
                 raise Exception('Maximum layer count for current size is exceeded!')
-            if side == 'top':
+            if side == Side.top:
                 # move right
                 dy0 = 1 if not self.compact and (center - pos_y) % 16 == 0 else 0
                 dy1 = 2 if not self.compact and (center - pos_y + 1) % 16 == 0 else 1
@@ -733,7 +769,7 @@ class AztecCode(object):
                 pos_x += 1
                 if num > max_num:
                     num = 2
-                    side = 'right'
+                    side = Side.right
                     pos_x -= 1
                     pos_y += 1
                 # skip reference grid
@@ -741,7 +777,7 @@ class AztecCode(object):
                     pos_x += 1
                 if not self.compact and (center - pos_y) % 16 == 0:
                     pos_y += 1
-            elif side == 'right':
+            elif side == Side.right:
                 # move down
                 dx0 = 1 if not self.compact and (center - pos_x) % 16 == 0 else 0
                 dx1 = 2 if not self.compact and (center - pos_x + 1) % 16 == 0 else 1
@@ -750,7 +786,7 @@ class AztecCode(object):
                 pos_y += 1
                 if num > max_num:
                     num = 2
-                    side = 'bottom'
+                    side = Side.bottom
                     pos_x -= 2
                     if not self.compact and (center - pos_x - 1) % 16 == 0:
                         pos_x -= 1
@@ -760,7 +796,7 @@ class AztecCode(object):
                     pos_y += 1
                 if not self.compact and (center - pos_x) % 16 == 0:
                     pos_x -= 1
-            elif side == 'bottom':
+            elif side == Side.bottom:
                 # move left
                 dy0 = 1 if not self.compact and (center - pos_y) % 16 == 0 else 0
                 dy1 = 2 if not self.compact and (center - pos_y + 1) % 16 == 0 else 1
@@ -769,7 +805,7 @@ class AztecCode(object):
                 pos_x -= 1
                 if num > max_num:
                     num = 2
-                    side = 'left'
+                    side = Side.left
                     pos_x += 1
                     pos_y -= 2
                     if not self.compact and (center - pos_y - 1) % 16 == 0:
@@ -779,7 +815,7 @@ class AztecCode(object):
                     pos_x -= 1
                 if not self.compact and (center - pos_y) % 16 == 0:
                     pos_y -= 1
-            elif side == 'left':
+            elif side == Side.left:
                 # move up
                 dx0 = 1 if not self.compact and (center - pos_x) % 16 == 0 else 0
                 dx1 = 2 if not self.compact and (center - pos_x - 1) % 16 == 0 else 1
@@ -788,7 +824,7 @@ class AztecCode(object):
                 pos_y -= 1
                 if num > max_num:
                     num = 2
-                    side = 'top'
+                    side = Side.top
                     layer_index += 1
                 # skip reference grid
                 if not self.compact and (center - pos_y) % 16 == 0:
@@ -800,7 +836,7 @@ class AztecCode(object):
         self.__add_finder_pattern()
         self.__add_orientation_marks()
         self.__add_reference_grid()
-        data_cw_count = self.__add_data(self.data)
+        data_cw_count = self.__add_data(self.data, self.encoding)
         self.__add_mode_info(data_cw_count)
 
 
@@ -808,6 +844,7 @@ def main(argv):
     if len(argv) not in (2, 3):
         print("usage: {} STRING_TO_ENCODE [IMAGE_FILE]".format(argv[0]))
         print("  Generate a 2D Aztec barcode and print it, or save to a file.")
+        raise SystemExit(1)
     data = argv[1]
     aztec_code = AztecCode(data)
     print('Aztec Code info: {0}x{0} {1}'.format(aztec_code.size, '(compact)' if aztec_code.compact else ''))
